;DHT22 Humidity/Temperature Sensor Library for Great Cow Basic

;Thomas Henry
;Version 1.0 -- 4/2/2014

;There is only one subroutine here. When called, it will
;return the relative humidity, Celsius and Faharenheit
;temperatures to one decimal place. The three numbers are
;treated as signed integers, scaled up by ten. For example,
;a return value of 657 for the relative humidity would be
;interpreted as 65.7%.

;Also returned is an error condition (a byte):
;0 = no error
;1 = no response from the sensor
;2 = bad checksum from the sensor

;Nine local bytes are consumed in the computations here,
;and seven bytes used for the output parameters. That's a
;grand total of sixteen bytes required when invoking this.

;The only constant required in the main calling program
;is that which sets the single data pin, e.g.,

;#define DHT22_pin Port A.0

;Note that whichever pin is employed, it must be capable
;of both input and output operation and must have a
;10k pull-up resistor.

;For reference, the pinout of the DHT22 is, looking from
;the front, (grill side):

;Pin 1: +5V
;Pin 2: DHT22_pin
;Pin 3: no connection
;Pin 4: ground

;This sensor transmits data via pulse code modulation at a
;fairly zippy rate, so the microcontroller clock should be
;at least 8 MHz. Observe that the DHT22 can not be
;polled more frequently than once every two seconds.

;----- Variables

dim DHT22_values(5)
dim DHT22_counter, DHT22_i, DHT22_byte as byte

;reuse a couple variables to save memory
dim DHT22_chksum alias DHT22_counter
dim DHT22_abs alias DHT22_counter

;----- Subroutine

sub readDHT22(out DHT22_rh as integer, out DHT22_cels as integer, out DHT22_fahr as integer, out DHT22_error as byte)

  ;----- request an update

  dir DHT22_pin out                     ;port is output now
  set DHT22_pin off                     ;go low
  wait 18 mS                            ;for 18 milliseconds
  set DHT22_pin on                      ;then go high
  wait 40 uS                            ;for 40 microseconds

  ;----- wait for an acknowledgment

  dir DHT22_pin in                      ;port is input now

  DHT22_counter = 0                     ;count by tens of microseconds
  do while DHT22_pin = off              ;until pin goes high
    wait 1 10uS
    DHT22_counter++
    if DHT22_counter > 9 then           ;should take 80 microseconds
      goto DHT22_noResponse             ;so must be a dud
    end if
  loop

  DHT22_counter = 0                     ;repeat, looking for a low now
  do while DHT22_pin = on
    wait 1 10uS
    DHT22_counter++
    if DHT22_counter > 9 then           ;should take 80 microseconds
      goto DHT22_noResponse             ;so must be a dud
    end if
  loop

  ;----- start of transmission

  for DHT22_i = 1 to 5                  ;five bytes to collect
    DHT22_byte = 0                      ;build up DHT22_byte here
    repeat 8                            ;8 bits to a byte
      DHT22_byte = 2 * DHT22_byte       ;shift left one place
      do while DHT22_pin = off          ;wait for start of pulse
      loop

      DHT22_counter = 0                 ;measure incoming pulse
      do while DHT22_pin = on
        wait 1 10uS                     ;count by tens of microseconds
        DHT22_counter++
      loop

      if DHT22_counter > 4 then         ;long pulse is a 1
        DHT22_byte++                    ;factor the bit in
      end if
    end repeat                          ;next bit

    DHT22_values(DHT22_i) = DHT22_byte  ;store complete byte
  next DHT22_i

  ;----- construct output values

  DHT22_chksum = DHT22_values(1)        ;compute the checksum
  for DHT22_i = 2 to 4
    DHT22_chksum = DHT22_chksum + DHT22_values(DHT22_i)
  next DHT22_i

  if DHT22_chksum <> DHT22_values(5) then
    DHT22_error = 2                     ;Error 2 means bad checksum
    goto DHT22_conclude
  else
    DHT22_error = 0                     ;Error 0 means all okay

    ;compute relative humidity scaled up by 10
    DHT22_rh = [integer]256 * DHT22_values(1) + DHT22_values(2)

    ;save the absolute value of the Celsius temperature
    DHT22_abs = DHT22_values(3) & 0x7F

    ;compute Celsius temperature scaled up by 10
    DHT22_cels = [integer]256 * DHT22_abs + DHT22_values(4)

    ;factor in the sign if needed
    if DHT22_values(3).7 = on then      ;a negative temp
      DHT22_cels = -DHT22_cels          ;make a true 2's complement
    end if

    DHT22_fahr = [integer]10*DHT22_cels       ;scaled up by 100 now
    DHT22_fahr = [integer]DHT22_fahr*9/5+3205 ;force rounding in the 100ths
    DHT22_fahr = [integer]DHT22_fahr/10       ;this is 10*Fahrenheit, rounded

    goto DHT22_conclude                 ;all done!
  end if

  DHT22_noResponse:
    DHT22_error = 1                     ;Error 1 means no response
  DHT22_conclude:
end sub